import { WDAG } from '../supporting/wdag.js';
import { Action } from '../display/action.js';
import { Actor } from '../display/actor.js';
import params from '../../params.js';
import { ReputationHolder } from '../reputation/reputation-holder.js';
import { displayNumber, EPSILON, INCINERATOR_ADDRESS } from '../../util.js';

const CITATION = 'citation';
const BALANCE = 'balance';

class Post extends Actor {
  constructor(forum, authorPublicKey, postContent) {
    const index = forum.posts.countVertices();
    const name = `Post${index + 1}`;
    super(name, forum.scene);
    this.id = postContent.id ?? name;
    this.authorPublicKey = authorPublicKey;
    this.value = 0;
    this.initialValue = 0;
    this.citations = postContent.citations;
    this.title = postContent.title;
    const leachingTotal = this.citations
      .filter(({ weight }) => weight < 0)
      .reduce((total, { weight }) => total += -weight, 0);
    const donationTotal = this.citations
      .filter(({ weight }) => weight > 0)
      .reduce((total, { weight }) => total += weight, 0);
    if (leachingTotal > params.revaluationLimit) {
      throw new Error('Post leaching total exceeds revaluation limit '
        + `(${leachingTotal} > ${params.revaluationLimit})`);
    }
    if (donationTotal > params.revaluationLimit) {
      throw new Error('Post donation total exceeds revaluation limit '
        + `(${donationTotal} > ${params.revaluationLimit})`);
    }
    if (this.citations.some(({ weight }) => Math.abs(weight) > params.revaluationLimit)) {
      throw new Error(`Each citation magnitude must not exceed revaluation limit ${params.revaluationLimit}`);
    }
  }

  getLabel() {
    return `${this.name}
    <table><tr>
      <td>initial</td>
      <td>${displayNumber(this.initialValue)}</td>
    </tr><tr>
      <td>value</td>
      <td>${displayNumber(this.value)}</td>
    </tr></table>`
      .replaceAll(/\n\s*/g, '');
  }
}

/**
 * Purpose:
 * - Forum: Maintain a directed, acyclic, graph of positively and negatively weighted citations.
 *   and the value accrued via each post and citation.
 */
export class Forum extends ReputationHolder {
  constructor(dao, name, scene) {
    super(name, scene);
    this.dao = dao;
    this.id = this.reputationPublicKey;
    this.posts = new WDAG(scene);
    this.actions = {
      propagate: new Action('propagate', scene),
      confirm: new Action('confirm', scene),
      transfer: new Action('transfer', scene),
    };
  }

  async addPost(authorId, postContent) {
    console.log('addPost', { authorId, postContent });
    const post = new Post(this, authorId, postContent);
    this.posts.addVertex(post.id, post, post.getLabel());
    for (const { postId: citedPostId, weight } of post.citations) {
      // Special case: Incinerator
      if (citedPostId === INCINERATOR_ADDRESS && !this.posts.getVertex(INCINERATOR_ADDRESS)) {
        this.posts.addVertex(INCINERATOR_ADDRESS, { name: 'Incinerator' }, 'Incinerator');
      }
      this.posts.addEdge(CITATION, post.id, citedPostId, weight);
    }
    return post;
  }

  getPost(postId) {
    return this.posts.getVertexData(postId);
  }

  getPosts() {
    return this.posts.getVerticesData();
  }

  async setPostValue(post, value) {
    post.value = value;
    await post.setValue('value', value);
    this.posts.setVertexLabel(post.id, post.getLabel());
  }

  getTotalValue() {
    return this.getPosts().reduce((total, { value }) => total += value, 0);
  }

  async onValidate({
    pool, postId, tokenId,
  }) {
    const initialValue = this.dao.reputation.valueOf(tokenId);
    const postVertex = this.posts.getVertex(postId);
    const post = postVertex.data;
    post.setStatus('Validated');
    post.initialValue = initialValue;
    this.posts.setVertexLabel(post.id, post.getLabel());

    // Store a reference to the reputation token associated with this post,
    // so that its value can be updated by future validated posts.
    post.tokenId = tokenId;

    const rewardsAccumulator = new Map();

    // Compute rewards
    await this.propagateValue(
      { to: postVertex, from: { data: pool } },
      { rewardsAccumulator, increment: initialValue },
    );

    // Apply computed rewards to update values of tokens
    for (const [id, value] of rewardsAccumulator) {
      if (value < 0) {
        this.dao.reputation.transferValueFrom(id, post.tokenId, -value);
      } else {
        this.dao.reputation.transferValueFrom(post.tokenId, id, value);
      }
    }

    // Transfer ownership of the minted/staked token, from the posts to the post author
    this.dao.reputation.transfer(this.id, post.authorPublicKey, post.tokenId);
    // const toActor = this.scene?.findActor((actor) => actor.reputationPublicKey === post.authorPublicKey);
    // const value = this.dao.reputation.valueOf(post.tokenId);
  }

  /**
   * @param {Edge} edge
   * @param {Object} opaqueData
   */
  async propagateValue(edge, {
    rewardsAccumulator, increment, depth = 0, initialNegative = false,
  }) {
    const postVertex = edge.to;
    const post = postVertex?.data;
    this.actions.propagate.log(edge.from.data, post, `(${increment})`);

    if (!!params.referenceChainLimit && depth > params.referenceChainLimit) {
      this.actions.propagate.log(
        edge.from.data,
        post,
        `referenceChainLimit (${params.referenceChainLimit}) reached`,
        null,
        '-x',
      );
      return increment;
    }

    console.log('propagateValue start', {
      from: edge.from.id ?? edge.from,
      to: edge.to.id,
      depth,
      value: post.value,
      increment,
      initialNegative,
    });

    const propagate = async (positive) => {
      let totalOutboundAmount = 0;
      const citationEdges = postVertex.getEdges(CITATION, true)
        .filter(({ weight }) => (positive ? weight > 0 : weight < 0));
      for (const citationEdge of citationEdges) {
        const { weight } = citationEdge;
        let outboundAmount = weight * increment;
        if (Math.abs(outboundAmount) > EPSILON) {
          const balanceToOutbound = this.posts.getEdgeWeight(BALANCE, citationEdge.from, citationEdge.to) ?? 0;
          let refundFromOutbound = 0;

          // Special case: Incineration.
          if (citationEdge.to.id === INCINERATOR_ADDRESS) {
            // Only a positive amount may be incinerated! Otherwise the sink could be used as a source.
            if (outboundAmount < 0) {
              throw new Error('Incinerator can only receive positive citations!');
            }
            // Reputation sent to the incinerator is burned! This means it is deducted from the sender,
            // without increasing the value of any other token.
            this.actions.propagate.log(citationEdge.from.data, { name: 'Incinerator' }, `(${increment})`);
          } else {
            // We need to ensure that we at most undo the prior effects of this post
            if (initialNegative) {
              outboundAmount = outboundAmount < 0
                ? Math.max(outboundAmount, -balanceToOutbound)
                : Math.min(outboundAmount, -balanceToOutbound);
            }

            // Recursively propagate reputation effects
            refundFromOutbound = await this.propagateValue(citationEdge, {
              rewardsAccumulator,
              increment: outboundAmount,
              depth: depth + 1,
              initialNegative: initialNegative || (depth === 0 && outboundAmount < 0),
            });

            // Any excess (negative) amount that could not be propagated,
            // i.e. because a cited post has been reduced to zero value,
            // is retained by the citing post.
            outboundAmount -= refundFromOutbound;
          }

          // Keep a record of the effect of the reputation transferred along this edge in the graph,
          // so that later, negative citations can be constrained to at most undo these effects.
          this.posts.setEdgeWeight(BALANCE, citationEdge.from, citationEdge.to, balanceToOutbound + outboundAmount);
          totalOutboundAmount += outboundAmount;

          this.actions.confirm.log(
            citationEdge.to.data,
            citationEdge.from.data,
            `(refund: ${displayNumber(refundFromOutbound)}, leach: ${outboundAmount * params.leachingValue})`,
            undefined,
            '-->>',
          );
        }
      }
      return totalOutboundAmount;
    };

    // First, leach value via negative citations
    const totalLeachingAmount = await propagate(false);
    increment -= totalLeachingAmount * params.leachingValue;

    // Now propagate value via positive citations
    const totalDonationAmount = await propagate(true);
    increment -= totalDonationAmount * params.leachingValue;

    // Apply the remaining increment to the present post
    const rawNewValue = post.value + increment;
    const newValue = Math.max(0, rawNewValue);
    const appliedIncrement = newValue - post.value;
    const refundToInbound = increment - appliedIncrement;

    console.log('propagateValue end', {
      depth,
      increment,
      rawNewValue,
      newValue,
      appliedIncrement,
      refundToInbound,
    });

    // Award reputation to post author
    rewardsAccumulator.set(post.tokenId, appliedIncrement);

    // Increment the value of the post
    await this.setPostValue(post, newValue);

    return refundToInbound;
  }
}
